#include "image_math.h"
#include "global.h"
#include "bip.h"

#include <cmath>
#include <iostream>
using namespace std;
using namespace stdi;

#define A_SCALE 57.2958

namespace { // unnamed namespace

  void checkPercentage(float p)
  {
    if (p<0 || p>1) {
      cerr << "In float image_math::meanOfTopByPercentage(vector<float> img, float p)" << endl;
      cerr << "Error: percentage must be a value in [0, 1]" << endl;
      exit(2);
    }
  }
  


} // unnamed namespace



PointArray image_math::findZeros(const Image &I)
{
  ByteImage J = bip::threshold(I, 0);
  J = bip::edge(J);
  const int nx = I.nx();
  const int ny = I.ny();
  int x, y;

  PointArray P;
  for(x=0; x<nx; x++)
    for(y=0; y<ny; y++)
      if (J(x,y)==Foreground)
	P.push_back(Point(x, y));
  return P;
}


PointArray image_math::findZeros(const Image &Ix, const Image &Iy)
{
  if (!Ix.equalSize(Iy)) return PointArray();

  ByteImage Jx = bip::threshold(Ix, 0);
  Jx = bip::edge(Jx);
  ByteImage Jy = bip::threshold(Iy, 0);
  Jy = bip::edge(Jy);

  const int nx = Ix.nx();
  const int ny = Ix.ny();
  int x, y;

  PointArray P;
  for(x=0; x<nx; x++)
    for(y=0; y<ny; y++)
      if (Jx(x,y)==Foreground && Jy(x,y)==Foreground)
	P.push_back(Point(x, y));
  return P;
}


float image_math::norm(float a, float b)
{
  return sqrt(a*a+b*b);
}

float image_math::norm2(float a, float b)
{
  return a*a+b*b;
}

float image_math::angle(float dx, float dy)
{
  if (dx == 0 && dy == 0) return 0;  
  return  A_SCALE * atan2(dx,dy);
}

byte image_math::direction(float dd)
{ 
  if (dd < 0) {
    if (dd >= -22.5) return 0;
    else if (dd >= -67.5) return 7;     /* 7 */
    else if (dd >= -112.5) return 6;    /* 6 */
    else if (dd >= -157.5) return 5;    /* 5 */
    else if (dd >= -180) return 4;      /* 4 */
  } else {
    if (dd <= 22.5) return 0;
    else if (dd <= 67.5) return 1;
    else if (dd <= 112.5) return 2;
    else if (dd <= 157.5) return 3;
    else return 4;
  }
  
}

byte image_math::direction(float a, float b)
{
  return  image_math::direction(image_math::angle(a, b));
}


float image_math::topByPixelNumber(const Image &img, int numpix)
{
  if (img.isNull()) {
    cerr << "In float topByPercentage(const Image &img, int numpix)" << endl;
    cerr << "Warning: img is null, return 0." << endl; 
    return 0;
  }
  const int npix = img.npix();
  if (numpix<0 || numpix>=npix) {
    cerr << "In float topByPercentage(const Image &img, int numpix)" << endl;
    cerr << "Error: numpix must be in [0, img.npix()-1]" << endl; 
    exit(2);
  }

  const int minv = (int) img.fmin(); 
  const int maxv = (int) img.fmax(); 
  if (minv==maxv) return minv;

  const int len = maxv-minv+1;
  vector<int> h(len, 0);
  int k;
  for(int i=0; i<npix; i++) {
    k = (int) (img.pixel(i) - minv);
    if (k<0) k = 0;
    if (k>=len) k = len-1;
    h[k]++;
  } 

  int m = 0;
  for(int i=len-1; i>=0; i--) {
    m += h[i];
    if (m>=numpix) return i+minv;
  }
  return minv;
}

float image_math::topByPercentage(const Image &img, float p)
{  
  checkPercentage(p);
  return image_math::topByPixelNumber( img, img.npix()*p );
}

float image_math::meanOfTopByPercentage(const Image &img, float p)
{

  checkPercentage(p);
  if (img.isNull()) {
    cerr << "In float topByPercentage(const Image &img, float p)" << endl;
    cerr << "Warning: img is null, return 0." << endl; 
    return 0;
  }

  const int minv = (int) img.fmin(); 
  const int maxv = (int) img.fmax(); 
  if (minv==maxv) return minv;
  
  const int len = maxv-minv+1;
  vector<int> h(len, 0);
  const int npix = img.npix();
  int k;
  for(int i=0; i<npix; i++) {
    k = (int) (img.pixel(i) - minv);
    if (k<0) k = 0;
    if (k>=len) k = len-1;
    h[k]++;
  } 

  int m = 0;
  int s = 0;
  const int numpix = (int)(p*npix);
  if (numpix<=0) return maxv;
  for(int i=len-1; i>=0; i--) {
    m += h[i];
    s += h[i]*(i+minv);
    if (m>=numpix) break;
  }
  if (m>0) return s/(float)m;
  reportBugInfo("float image_math::meanOfTopByPercentage(const Image &img, float p)");
}


float image_math::meanOfTopByPercentage(vector<float> img, float p)
{
  if (fabs(p-1)<0.01) 
    return image_math::mean(img);
   
  checkPercentage(p);
  if (img.empty()) return 0;
  const int minv = (int) image_math::minimum(img); 
  const int maxv = (int) image_math::maximum(img); 
  if (minv==maxv) return minv;
  
  const int len = maxv-minv+1;
  vector<int> h(len, 0);
  const int npix = img.size();
  int k;
  for(int i=0; i<npix; i++) {
    k = (int)(img[i] - minv);
    if (k<0) k = 0;
    if (k>=len) k = len-1;
    h[k]++;
  } 
  

  const int numpix = (int)(npix * p);
  if (numpix<=0) return maxv;

  int m = 0;
  int s = 0;
  for(int i=len-1; i>=0; i--) {
    m += h[i];
    s += h[i]*(i+minv);
    if (m>=numpix) break;
  }

  if (m>0) return s/(float)m;
  reportBugInfo("float image_math::meanOfTopByPercentage(vector<float> img, float p)");
}
