#include "image_io.h"
#include "image.h"
#include "fnp.h"
#include "tiff.h"
#include "tiffio.h"

using namespace stdi;
using namespace image_io;


#include <cstdio>
#include <cstring>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <ctime>
using namespace std;


template<class T>
void reverse_byte_order(T* v, int len)
{
  const int n = sizeof(T);
  const int n_1 = n-1;
  int i, k;
  T* x = v;
  T y;
  byte* px, *py;
  py = (byte*) &y;
  for(k=0; k<len; k++, x++) {
    y = *x;
    px = (byte*) x;
    for(i=0; i<n; i++) {
      px[i] = py[n_1-i];
    } 
  }
}



void reverse_byte_order(const Image& I)
{
  switch (I.format()) {
      case BYTE:
	  return;
      case SHORT:
	  reverse_byte_order<short>((short*)I.data(), I.npix());
	  return;
     case USHORT:
	  reverse_byte_order<unsigned short>((unsigned short*)I.data(), I.npix());
	  return;
      case INT:
	  reverse_byte_order<int>((int*)I.data(), I.npix());
	  return;
      case UINT:
	  reverse_byte_order<unsigned int>((unsigned int*)I.data(), I.npix());
	  return;
      case FLOAT:
	  reverse_byte_order<float>((float*)I.data(), I.npix());
	  return;
      case RGB:
	  return;
      case SHORTRGB:
	  reverse_byte_order<short>((short*)I.data(), I.npix()*3);
	  return;
      case INTRGB:
	  reverse_byte_order<int>((int*)I.data(), I.npix()*3);
	  return;
      case FLOATRGB:
	  reverse_byte_order<float>((float*)I.data(), I.npix()*3);
	  return;
      case RGBA:
	  return;
      default:
	  ;
  }
}


void adjust_byte_order_for_read(const Image& I, bool big_endian)
{
#ifdef LINUX_VER
  if (big_endian) 
    reverse_byte_order(I);
#else
  if (!big_endian) 
    reverse_byte_order(I);
#endif
  
}

void adjust_byte_order_for_write(const Image& I)
{
#ifdef LINUX_VER
  reverse_byte_order(I);
#endif
}

// Local functions in unnamed namespace
namespace {

// Functions for PGM io utilities
#define  exit_error(IFB,IFC) { fprintf(stderr,IFB,IFC); exit(1); }
#define  FTOI(a) ( (a) < 0 ? ((int)(a-0.5)) : ((int)(a+0.5)) )

int getint( FILE *fd)
{
  int c, i;
  char dummy[10000];
  
  c = getc(fd);
  while (1) /* find next integer */
    {
    if (c=='#')    /* if we're at a comment, read to end of line */
      fgets(dummy,9000,fd);
    if (c==EOF)
      exit_error("Image %s not binary PGM.\n","is");
    if (c>='0' && c<='9')
      break;   /* found what we were looking for */
    c = getc(fd);
    }

  /* we're at the start of a number, continue until we hit a non-number */
  i = 0;
  while (1) {
    i = (i*10) + (c - '0');
    c = getc(fd);
    if (c==EOF) return (i);
    if (c<'0' || c>'9') break;
  }

  return (i);
}


void get_image(  const char *filename, 
		 byte  **in,
		 int  *x_size, int  *y_size)
{
  FILE  *fd;
  char header [256];
  int npix;

#ifdef GNUDOS
  if ((fd=fopen(filename,"rb")) == NULL)
#else
    if ((fd=fopen(filename,"r")) == NULL)
#endif
      exit_error("Can't input image %s.\n",filename);
  
  header[0]=fgetc(fd);
  header[1]=fgetc(fd);
  if(!(header[0]=='P' && header[1]=='5'))
    exit_error("Image %s does not have binary PGM header.\n",filename);
  
  *x_size = getint(fd);
  *y_size = getint(fd);
  getint(fd);

  npix = *x_size * *y_size;  
  *in = (byte *) malloc(npix);
  if (! *in ) {
    fprintf(stderr, "out of memory\n");
    exit(1);
  }

  if (fread(*in, 1, npix, fd) != npix)
    exit_error("Image %s is wrong size.\n",filename);
  
  fclose(fd);
}

void put_image(const char *filename,
	       byte    *in,
	       int  x_size,
	       int  y_size)
{
  FILE  *fd;
  
#ifdef GNUDOS
  if ((fd=fopen(filename,"wb")) == NULL) 
#else
    if ((fd=fopen(filename,"w")) == NULL) 
#endif
      exit_error("Can't output image%s.\n",filename);
  
  fprintf(fd,"P5\n");
  fprintf(fd,"%d %d\n", x_size, y_size);
  fprintf(fd,"255\n");
  
  if (fwrite(in, 1, x_size*y_size, fd) != x_size*y_size)
    exit_error("Can't write image %s.\n",filename);
  
  fclose(fd);
}

template <class T>
void __image_format_conv__(const Image &oimg, image<T> &nimg)
{
  if (oimg.isNull()) return;
  const int npix = oimg.npix();
  nimg = image<T>(oimg.nx(), oimg.ny(), oimg.nz());
  for(int i=0; i<npix; i++)
    nimg.set(i, (T) oimg.pixel(i));
}



} // End of unnamed namespace


Image *image_io::readPGM(const string &_filename)
{
  const char *filename = _filename.c_str();
  byte  *in;
  int  x_size, y_size,k;  
  get_image(filename,&in,&x_size,&y_size);
  ByteImage  *im = new ByteImage(x_size, y_size);
  int num = x_size*y_size;
  for (k = 0; k < num; k++)  im->set(k, (byte) in[k]);
  free(in);
  return im;
}


void image_io::writeUIF(const string &_filename, const Image &img)
{
  const char *filename = _filename.c_str();

  int nx = img.nx();
  int ny = img.ny();
  int nz = img.nz();
  int pixel_size = img.pixelSize();
  PixelFormat pixel_format = img.format();
  
  ofstream  outimage(filename);
  
  if(!outimage) {
    cerr << "In image_io::writeUIF" << endl;
    cerr << "Can not open file" << filename << "\n";
    cerr << "Ignored." << endl;
    return;
  }
  
  outimage.write((char *)(&nx), sizeof(int));
  outimage.write((char *)(&ny), sizeof(int));
  outimage.write((char *)(&nz), sizeof(int));
  outimage.write((char *)(&pixel_size), sizeof(int));
  outimage.write((char *)(&pixel_format), sizeof(PixelFormat));  
  outimage.write((char *) img.data(), img.npix() * img.pixelSize());
  
  outimage.close();
  
}


Image *image_io::readUIF(const string &_filename)
{
  const char *filename = _filename.c_str();
  ifstream inimage(filename);
    
    if (!inimage){
      cerr << "In image_io::readUIF" << endl;
      cout <<"Can not open the file " << filename << "\n" ;
      return NULL;      
    }    
    
    int nx, ny, nz, pixel_size;
    PixelFormat pixel_format;
  
    inimage.read((char *)(&nx), sizeof(int));
    inimage.read((char *)(&ny), sizeof(int));
    inimage.read((char *)(&nz), sizeof(int));
    inimage.read((char *)(&pixel_size), sizeof(int));
    inimage.read((char *)(&pixel_format), sizeof(PixelFormat));    
  
    Image *img = NULL;
    
    switch (pixel_format) {
    case BYTE:
      img = new ByteImage(nx, ny, nz); break;
    case SHORT:
      img = new ByteImage(nx, ny, nz); break;
    case INT:
      img = new ByteImage(nx, ny, nz); break;
    case FLOAT:
      img =  new ByteImage(nx, ny, nz); break;
    default: 
      ;
    }

    if (!img) return NULL;

    inimage.read((char*) img->data(), img->npix() * img->pixelSize() );
    inimage.close();

    return  img;

}


void image_io::writePGM(const string &_filename, const Image &img)
{

  const char *filename = _filename.c_str();

  if (img.isNull()) {
    cerr << "In image_io::writePGM" << endl;
    cerr << "Warning: cannot write null image" << endl;
    return;
  }
  
  if (img.format()==BYTE) {
    put_image(filename, (byte *)img.data(), img.nx(), img.ny());
    return;
  }
  
  ByteImage a(img.nx(), img.ny());

  const int nxy(img.nx() * img.ny());  
  float minv = img.pixel(0);
  float maxv = img.pixel(0);
  for(int k=1; k<nxy; k++) {
    minv = min(minv, img.pixel(k)); 
    maxv = max(maxv, img.pixel(k)); 
  }

  if (minv>=maxv) 
    put_image(filename, (byte *)a.data(), img.nx(), img.ny());
  else 
    for(int k=0; k<nxy; k++)
      a.set(k, 255*(img.pixel(k)-minv)/(maxv-minv));
  
  put_image(filename, (byte *) a.data(), img.nx(), img.ny());
  
}


Image *image_io::readIHT(const string &_filename)
{
  const char *filename = _filename.c_str();

    ifstream    inimage(filename);
    if (!inimage){
      cout <<"Can not open the file " << filename << "\n" ;
	return NULL;        
    }    

    int row;
    int col;
    int frow;
    int fcol;
    int orow;
    int ocol;
    int nfr;
    int num;
    PixelFormat pf;
  
    inimage.read((char *)(&row),    sizeof(int));
    inimage.read((char *)(&col),    sizeof(int));
    inimage.read((char *)(&frow),   sizeof(int));
    inimage.read((char *)(&fcol),   sizeof(int));
    inimage.read((char *)(&orow),   sizeof(int));
    inimage.read((char *)(&ocol),   sizeof(int));
    inimage.read((char *)(&pf),     sizeof(PixelFormat));
    inimage.read((char *)(&nfr),    sizeof(int));
    inimage.read((char *)(&num),    sizeof(int));
    inimage.seekg(sizeof(byte *), ios::cur);

     
    Image *img = NULL;
    switch (pf) {
    case BYTE:
      img = new ByteImage(col, row, nfr); break;
    case SHORT:
      img =  new ShortImage(col, row, nfr); break;
    case INT:
      img = new IntImage(col, row, nfr); break;
    case FLOAT:
      img =  new FloatImage(col, row, nfr); break;
    default: 
      ;
    }

    if (!img) return NULL;    
    inimage.read((char*) img->data(), img->pixelSize()*img->npix());
    inimage.close();
    
    adjust_byte_order_for_read(*img, true);
    return img;
}



void image_io::writeIHT(const string &_filename, const Image& img )
{
  
  const char *filename = _filename.c_str();

  ofstream    outimage(filename);
  if(!outimage) {
    cout<< "Can not open file" << filename << "\n";
    //    exit ( 1 );
    return;
  }
  
  int  row = img.ny();
  int  col = img.nx();
  int  frow = 0;
  int  fcol = 0;
  int  orow = row;
  int  ocol = col;
  PixelFormat pf = img.format();
  int  nfr  = img.nz();
  int  num  = img.npix();
  byte *data = (byte *) img.data();

  
  outimage.write((char *)(&row),    sizeof(int));
  outimage.write((char *)(&col),    sizeof(int));
  outimage.write((char *)(&fcol),   sizeof(int));
  outimage.write((char *)(&frow),   sizeof(int));
  outimage.write((char *)(&orow),   sizeof(int));
  outimage.write((char *)(&ocol),   sizeof(int));
  outimage.write((char *)(&pf),     sizeof(PixelFormat));
  outimage.write((char *)(&nfr),    sizeof(int));
  outimage.write((char *)(&num),    sizeof(int));
  outimage.write((char *)(&data),    sizeof(byte *));

  adjust_byte_order_for_write(img);
  outimage.write((char *) data, img.npix()*img.pixelSize());
  adjust_byte_order_for_write(img);

  outimage.close();

}

Image *image_io::readImage(const string &_filename ) 
{

  const char *filename = _filename.c_str();

  string _ext = fnp::extension(filename);
  const char * ext = _ext.c_str();

  if ((!strcasecmp(ext, "pgm")) || (!strcmp(ext,"PGM")))
    return image_io::readPGM(filename);
  if ( (!strcasecmp(ext, "iht")) || (!strcmp(ext,"IHT") ))
    return image_io::readIHT(filename);
  if ((!strcmp(ext, "ics")) || (!strcmp(ext,"ICS") ))
    return image_io::readICS(filename);
 
  return NULL;
}


void image_io::writeImage(const string &_filename, const Image &img) 
{
  using namespace fnp;

  const char *filename = _filename.c_str();
  string _ext = extension(clean(filename));
  const char * ext = _ext.c_str();
  
  if (!strcasecmp(ext, "pgm"))
    return image_io::writePGM(filename, img);
  if (!strcasecmp(ext, "iht"))
    return image_io::writeIHT(filename, img);
  if (!strcmp(ext, "ics"))
    return image_io::writeICS(filename, img);
}


ByteImage image_io::convertToByteImage(const Image &img, bool forced_scaling)
{
  if (img.isNull()) return ByteImage();
  if (img.format()==BYTE && !forced_scaling) 
    return * ( (ByteImage *) &img );
  const int npix ( img.npix() );
  ByteImage a(img.nx(), img.ny(), img.nz());

  float val = img.pixel(0);
  float minv = val;
  float maxv = val;
  for(int k=1; k<npix; k++) {
    val = img.pixel(k); 
    if (val<minv) minv = val;
    else if (val>maxv) maxv = val;
  }
  if (fabs(minv-maxv)<0.0001) return a;
  
  if (minv>=0 && maxv<=255 && !forced_scaling)
    for(int k=0; k<npix; k++)
      // a.set(k, byte(max(0, min(255, (int)img.pixel(k)))) );
      a.set(k, byte( (int)img.pixel(k) ));
  else
    for(int k=0; k<npix; k++)
      a.set(k, 255*(img.pixel(k)-minv)/(maxv-minv));
  
  return a;
}

extern "C"
int parse_ics_header_file(const char *ics,
			  char **filename, char **fmt, 
			  int *nx, int *ny, int *nz, int*nc, int *nt,
			  int* is_big_endian);


Image *image_io::readICS(const string &_ics_filename)
{
    const char *ics_filename = _ics_filename.c_str();
    string _ext = fnp::extension(ics_filename); //sree
    const char * ext = _ext.c_str();	     //sree
    
    char *fname, *fmt;
    int nx, ny, nz,nc,nt,nzval;
    int is_big_endian = 1;
    if (!parse_ics_header_file(ics_filename, &fname, &fmt, 
			       &nx, &ny, &nzval, &nc, &nt, &is_big_endian ) ) {
	
	cerr << "In image_io::readICS" << endl;
	cerr << "   Warning: failed tp parse ICS file" << endl;
	cerr << "   Null image pointer returned" << endl;
	cout << "error in parse_ics_header_file !"<<endl;
	free(fname); free(fmt); return NULL;
	
    }
    
    nz = nzval * nc * nt; 
   
    cout << "nx:"<<nx<<endl;
    cout << "ny:"<<ny<<endl;
    cout << "nz:"<<nzval<<endl;
    cout << "nc:"<<nc<<endl;
    cout << "nt:"<<nt<<endl;
   
    if ( !fmt || !fname || nx<=0 || ny<=0 || nz<=0)  {
	free(fname); free(fmt); return NULL; 
    }
      
    Image *img = 0; 
    cout<<"inreadICS format:"<<fmt<<endl;
    if (!strcasecmp(fmt, "BYTE")) 
	img = new ByteImage(nx, ny, nzval ,nc,nt);
    else  if (!strcasecmp(fmt, "USHORT"))
	img = new UShortImage(nx, ny, nzval,nc,nt ); 
    else  if (!strcasecmp(fmt, "SHORT")) 
	img = new ShortImage(nx, ny, nzval,nc,nt); 
    else  if (!strcasecmp(fmt, "UINT"))
	img = new UIntImage(nx, ny, nzval,nc,nt );
    else if (!strcasecmp(fmt, "INT")) 
    //img = new IntImage(ny, nx, nz,nc,nt,nzval);
	img = new UIntImage(nx, ny, nzval,nc,nt);
    else if (!strcasecmp(fmt, "FLOAT")) 
	img = new FloatImage(nx, ny, nzval,nc,nt);
    else if (!strcasecmp(fmt, "RGB")) 
	img = new RgbImage(nx, ny, nzval,nc,nt);
    else if (!strcasecmp(fmt, "INTRGB")) 
	img = new IntRgbImage(nx, ny, nzval,nc,nt);
    else if (!strcasecmp(fmt, "SHORTRGB")) 
	img = new ShortRgbImage(nx, ny, nzval,nc,nt);    
    else if (!strcasecmp(fmt, "FLOATRGB")) 
	img = new FloatRgbImage(nx, ny, nzval,nc,nt);
    else if (!strcasecmp(fmt, "RGBA")) 
	img = new RgbaImage(nx, ny, nzval,nc,nt);

    cout<<"Format "<<fmt<<endl;

    string ids_filename;
    if(!strcmp(ext,"ics"))
	ids_filename= fnp::root(ics_filename) + ".ids";
    else	 
	ids_filename= fnp::root(ics_filename) + ".IDS";
    ifstream inimage(ids_filename.c_str(), ios::binary);
    if (!inimage){
	cerr << "In image_io::readICS" << endl;
	cout << "   Error: can not open the file " << ids_filename << "\n" ;
	cerr << "   Null image pointer returned" << endl;
	free(fname); free(fmt); return 0;
    }    
    cout<<"Pixel Size"<<img->pixelSize()<<" Num pix :"<<img->npix()<<endl; 
    if(img)  
	if (!inimage.read((char*) img->data(), img->pixelSize()*img->npix())) {
	    cerr << "In image_io::readICS" << endl;
	    cerr << "   Warning: failed to read IDS file" << endl;    
	    delete img; 
	    cerr << "   Null image pointer returned" << endl;
	    img = 0;
	}
    
    free(fname); free(fmt);
    inimage.close();
    
    adjust_byte_order_for_read(*img, is_big_endian);
    
    return img;
}

namespace {


  void appendItem(string &s, const string &item)
  {
    if (!item.empty())
      s += "\t" + item;
  }

}


void image_io::writeICS( const string &__filename, const Image &I)
{

    const char *_filename = __filename.c_str();
    
    using namespace fnp;
    
    string filename = clean(_filename);
    string icsfile, idsfile;
    if ( (extension(filename)).compare("ics") ) {
	icsfile = string(filename)+".ics";
	idsfile = string(filename)+".ids";
    } else {    
	icsfile = string(filename);
	idsfile = root(filename)+".ids";
    }

    ofstream  outics(icsfile.c_str());  
    if(!outics) {
	cerr << "In image_io::writeICS" << endl;
	cerr << "   Can not open file" << filename << "\n";
	cerr << "   Ignored." << endl;
	return;
    }
    
    outics << "\t" << endl;
    outics << "ics_version\t1.0" << endl;
    outics << "filename\t" << tail(root(icsfile)) << endl;
    
    outics << "layout\tparameters\t6" << endl;
    outics << "layout\torder\tbits\tx\ty\tz\tc\tt"  << endl;
    outics << "layout\tsizes\t"<< 8*I.pixelSize() << "\t" <<
	I.nx() << "\t" << I.ny() << "\t" << I.nzval() << "\t" << I.nc()<< "\t" << I.nt()<< endl;
        
    outics << "layout\tcoordinates\tvideo" << endl;
    outics << "layout\tsignificant_bits\t" << 8*I.pixelSize() << endl;
    outics << "representation\tformat\t";
    
    switch( I.format() ){
	
	case UINT:
	    outics << "unint" << endl;
	    break;
	case USHORT:
	    outics << "unshort" << endl;
	    break;
	case BYTE:
	    outics << "integer" << endl;
	    break;
	case INT:
	    outics << "integer" << endl;
	    break;
	case SHORT:
	    outics << "integer" << endl;
	    break;
	case FLOAT:
	    outics << "float" << endl;
	    break;
	case RGB:
	    outics << "rgb" << endl;
	    break;
	case SHORTRGB:
	    outics << "srgb" << endl;
	    break;
	case INTRGB:
	    outics << "irgb" << endl;
	    break;
	case FLOATRGB:
	    outics << "frgb" << endl;
	    break;
	case RGBA:
	    outics << "rgba" << endl;
	    break;
	default:
	    outics << "unknown" << endl;
    }

    if( I.format() == UINT || I.format() == BYTE || I.format() == USHORT || I.format() == RGB || I.format() == RGBA )
	outics << "representation\tsign\tunsigned"<< endl;
    else
	outics << "representation\tsign\tsigned"<< endl;
    outics << "representation\tcompression\tuncompressed" << endl;
    outics << "representation\tbyte_order\t2\t1"<< endl;
    if (I.nzval()>1)outics << "representation\tSCIL_TYPE\tg3d" << endl;
    else  outics << "representation\tSCIL_TYPE\tg2d" << endl;
    outics << "parameter\tex_filter\tUNDEFINED" << endl;
    outics << "parameter\texp_time\tUNDEFINED" << endl;
    outics << "history\tdate\tUNDEFINED" << endl;
    outics << "history\tUTC_seconds\tUNDEFINED" << endl;
    
    outics.close();
    
    
    ofstream    outimage(idsfile.c_str());
    if(!outimage) {
	cerr << "In image_io::writeICS" << endl;
	cout<< "   Can not open file" << idsfile << endl;
	cerr << "  In image_io::writeICS" << endl;
	return;
    }
    
    
    adjust_byte_order_for_write(I);
    if (!outimage.write((char *) I.data(), I.npix()*I.pixelSize())) {
	cerr << "In image_io::writeICS" << endl;
	cerr << "   Error in writing " << idsfile<< endl;
    }
    adjust_byte_order_for_write(I);
    outimage.close();
}



// Convert to SHORT image
ShortImage image_io::convertToShortImage(const Image &img)
{
  ShortImage nimg;
  __image_format_conv__(img, nimg);
  return nimg;
}


// Convert to INT image
IntImage image_io::convertToIntImage(const Image &img)
{
  IntImage nimg;
  __image_format_conv__(img, nimg);
  return nimg;
}

UIntImage image_io::convertToUIntImage(const Image &img) //sree
{
  UIntImage nimg;
  __image_format_conv__(img, nimg);
  return nimg;
	
}

// Convert to FLOAT image
FloatImage image_io::convertToFloatImage(const Image &img)
{
  FloatImage nimg;
  __image_format_conv__(img, nimg);
  return nimg;
}


IcsInfo::IcsInfo(const string &_base_dataset_name)
  : base_dataset_name(_base_dataset_name),
    history_date(),
    history_utc_seconds(),
    param_ex_filter(),
    param_exp_time(),
    channel(),
    annotation(),
    scope(),
    sample(),
    _UTCseconds(0),
    other_properties()
{
}

IcsInfo::IcsInfo(const IcsInfo & orig)
  : base_dataset_name(orig.base_dataset_name),
    history_date(orig.history_date),
    history_utc_seconds(orig.history_utc_seconds),
    param_ex_filter(orig.param_ex_filter),
    param_exp_time(orig.param_exp_time),
    channel(orig.channel),
    annotation(orig.annotation),
    scope(orig.scope),
    sample(orig.sample),
    _UTCseconds(orig._UTCseconds),
    other_properties(orig.other_properties)
{
}

IcsInfo & IcsInfo::operator= (const IcsInfo &orig)
{
  this->base_dataset_name = orig.base_dataset_name;
  this->history_date = orig.history_date;
  this->history_utc_seconds = orig.history_utc_seconds;
  this->param_ex_filter = orig.param_ex_filter;
  this->param_exp_time = orig.param_exp_time;
  this->channel = orig.channel;
  this->annotation = orig.annotation;
  this->scope = orig.scope;
  this->sample = orig.sample;
  this->_UTCseconds = orig._UTCseconds;
  this->other_properties = orig.other_properties;
  return *this;
}

IcsInfo::~IcsInfo()
{
}

void IcsInfo::setWithChannelInfo(const string& ex_freq, 
				 const string& exp_time,
				 char filenamechannelchar) 
{
  param_ex_filter =  "parameter\tex_filter";
  appendItem(param_ex_filter, ex_freq);

  param_exp_time = "parameter\texp_time";
  appendItem(param_exp_time, exp_time);

  channel = string(" ");
  channel[0] = filenamechannelchar;
}


string IcsInfo::datasetName() const
{
  if (channel.empty()) return base_dataset_name;
  return base_dataset_name + "_" + channel;
}

long IcsInfo::getUTCseconds() const
{
  return  _UTCseconds;
}

void IcsInfo::setWithCurrentTime()
{
  char __date[30]; 
  char __seconds[32];
  time_t caltime;
  struct tm *tm;
  caltime= time(NULL); 
  _UTCseconds = caltime;

  tm=localtime(&caltime);
  strftime(__date,30,"%a %b %d %X %Z %Y",tm);
  sprintf(__seconds,"%d",caltime);
  
  history_date = string("history") + "\t" +
    string("date") + "\t" +
    string (__date);
  
  history_utc_seconds = string("history") + "\t" +
    string("UTC_seconds") + "\t" + 
    string (__seconds);

}

void IcsInfo::addProperty(const vector<string> &a)
{
  if (a.empty()) return;
  string s = a[0];
  for(int i=1; i<a.size(); i++)
    if (!a[i].empty())
      s += "\t" + a[i];
  other_properties.push_back(s);
}


void IcsInfo::addParagraph(const string &a)
{
  other_text += a;
}

void IcsInfo::annotate(const string & a)
{
  annotation = string();
  if (a.empty()) 
    return;

  const string h("history\tAnnotation\t");
  annotation.reserve(a.size() + h.size()*4);

  for(int i=0; i<a.size(); i++) {
    if ( !i || a[i-1]=='\n') {      
      annotation += h;
    }
    annotation += a[i];
  }
}

void IcsInfo::setScope(const string &_scope)
{
  scope = string("history\tScope");
  appendItem(scope, _scope);
}

void IcsInfo::setSample(const string &dish_type, const string &cell_type,
	       const string &media_name, float media_vol,
	       float media_con, const string &oil_type)
{
  sample = "Sample";
  appendItem(sample, dish_type);
  appendItem(sample, cell_type);
  appendItem(sample, media_name);

  char buf[256];
  sprintf(buf,"%f\0", media_vol);
  appendItem(sample,  string(buf));
  sprintf(buf,"%f\0", media_con);
  appendItem(sample,  string(buf));
  appendItem(sample,  oil_type);
}


string IcsInfo::channelName() const
{
  return channel;
}

ostream& image_io::operator<< (ostream &stream, const IcsInfo & info)
{
  // Write parameters
  if (!info.param_ex_filter.empty())
    stream << info.param_ex_filter << endl;
  //  else 
  //  stream << "parameter\tex_filter\tUNDEFINED" << endl;

  if (!info.param_exp_time.empty())
    stream << info.param_exp_time << endl;
  // else 
  //  stream << "parameter\texp_time\tUNDEFINED" << endl;
  
  // Write history
  // Date
  if (!info.history_date.empty()) 
    stream << info.history_date << endl;
  else  
    stream << "history\tdate\tUNDEFINED" << endl;
  
  // UTC_seconds
  if (!info.history_utc_seconds.empty()) 
    stream << info.history_utc_seconds << endl;
  else
    stream << "history\tUTC_seconds\tUNDEFINED" << endl;
  
  // Scope
  if (!info.scope.empty()) 
    stream  << info.scope << endl;
  else 
    stream << "history\tScope\tUNDEFINED" << endl;

  // Sample
  if (!info.sample.empty()) 
    stream  << info.sample << endl;

  // Other properties
  for(int i=0; i<info.other_properties.size(); i++) 
    if (!info.other_properties[i].empty())
      stream << info.other_properties[i] << endl;

  if (!info.other_text.empty()) {
    stream << info.other_text;
  }

  // Annotation
  if (!info.annotation.empty()) 
    stream << info.annotation;
  else 
    stream << "history\tAnnotation\tNONE";

  return stream;
}


void image_io::writeICS(const IcsInfo &info, const Image &I)
{
  if (info.datasetName().empty()) return;

  using namespace fnp;
  string filename = clean(info.datasetName());

  string icsfile, idsfile;
  if ( (extension(filename)).compare("ics") ) {
    icsfile = string(filename)+".ics";
    idsfile = string(filename)+".ids";
  } else {    
    icsfile = string(filename);
    idsfile = root(filename)+".ids";
  }
  
  //cout << icsfile << endl;
  //cout << idsfile << endl;
  ofstream  outics(icsfile.c_str());  
  if(!outics) {
    cerr << "In image_io::writeICS" << endl;
    cerr << "   Can not open file" << icsfile << "\n";
    cerr << "   Ignored." << endl;
    return;
  }
  
  outics << "\t" << endl;
  outics << "ics_version\t1.0" << endl;
  outics << "filename\t" << tail(root(icsfile)) << endl;

  if (I.nz()>1) {
    outics << "layout\tparameters\t4" << endl;
    outics << "layout\torder\tbits\tx\ty\tz"  << endl;
    outics << "layout\tsizes\t"<< 8*I.pixelSize() << "\t" <<
      I.ny() << "\t" << I.nx() << "\t" << I.nz() << endl;
  } else {
    outics << "layout\tparameters\t3" << endl;
    outics << "layout\torder\tbits\tx\ty"  << endl;
    outics << "layout\tsizes\t"<< 8*I.pixelSize() << "\t" <<
      I.ny() << "\t" << I.nx() << endl;
  }

  outics << "layout\tcoordinates\tvideo" << endl;
  outics << "layout\tsignificant_bits\t" << 8*I.pixelSize() << endl;
  outics << "representation\tformat\t";
  if (I.format()!=FLOAT) outics << "integer" << endl;
  else outics<< "float" << endl;
  outics << "representation\tsign\tsigned"<< endl;
  outics << "representation\tcompression\tuncompressed" << endl;
  outics << "representation\tbyte_order\t2\t1"<< endl;
  if (I.nz()>1)outics << "representation\tSCIL_TYPE\tg3d" << endl;
  else  outics << "representation\tSCIL_TYPE\tg2d" << endl;

  outics << info << endl;

  outics.close();  
  
  ofstream    outimage(idsfile.c_str());
  if(!outimage) {
    cerr << "In image_io::writeICS" << endl;
    cout<< "   Can not open file" << idsfile << endl;
    cerr << "  In image_io::writeICS" << endl;
    return;
  }

  adjust_byte_order_for_write(I);  
  if (!outimage.write((char *) I.data(), I.npix()*I.pixelSize())) {
    cerr << "In image_io::writeICS" << endl;
    cerr << "   Error in writing " << idsfile<< endl;
  }

  adjust_byte_order_for_write(I);

  outimage.close();

}






MultiChannelIcsInfo::MultiChannelIcsInfo(const string &_base_dataset_name)
  : base_dataset_name(_base_dataset_name),
    history_date(),
    history_utc_seconds(),
    param_ex_filter(),
    param_exp_time(),
    channel(),
    annotation(),
    scope(),
    sample(),
    _UTCseconds(0),
    other_properties()
{
}

MultiChannelIcsInfo::MultiChannelIcsInfo(const MultiChannelIcsInfo & orig)
  : base_dataset_name(orig.base_dataset_name),
    history_date(orig.history_date),
    history_utc_seconds(orig.history_utc_seconds),
    param_ex_filter(orig.param_ex_filter),
    param_exp_time(orig.param_exp_time),
    channel(orig.channel),
    annotation(orig.annotation),
    scope(orig.scope),
    sample(orig.sample),
    _UTCseconds(orig._UTCseconds),
    other_properties(orig.other_properties)
{
}

MultiChannelIcsInfo & MultiChannelIcsInfo::operator= (const MultiChannelIcsInfo &orig)
{
  this->base_dataset_name = orig.base_dataset_name;
  this->history_date = orig.history_date;
  this->history_utc_seconds = orig.history_utc_seconds;
  this->param_ex_filter = orig.param_ex_filter;
  this->param_exp_time = orig.param_exp_time;
  this->channel = orig.channel;
  this->annotation = orig.annotation;
  this->scope = orig.scope;
  this->sample = orig.sample;
  this->_UTCseconds = orig._UTCseconds;
  this->other_properties = orig.other_properties;
  return *this;
}

MultiChannelIcsInfo::~MultiChannelIcsInfo()
{
}

void MultiChannelIcsInfo::setWithChannelInfo(const string& ex_freq, 
				 const string& exp_time,
				 char filenamechannelchar) 
{

  param_ex_filter.push_back(  string("parameter") + "\t" + 
    string("ex_filter") + "\t" + ex_freq );

  param_exp_time.push_back(string("parameter")+ "\t" + 
    string("exp_time")+ "\t" + exp_time);

  string ch = " ";
  ch[0] = filenamechannelchar;
  channel.push_back(ch);
}

const string &MultiChannelIcsInfo::datasetName() const
{
  return base_dataset_name;
}



long MultiChannelIcsInfo::getUTCseconds() const
{
  return  _UTCseconds;
}

void MultiChannelIcsInfo::setWithCurrentTime()
{
  char __date[30]; // FIX : change to time_t and charsize?
  char __seconds[32]; //:)
  time_t caltime;
  struct tm *tm;
  caltime= time(NULL); 
  _UTCseconds = caltime;

  tm=localtime(&caltime);
  strftime(__date,30,"%a %b %d %X %Z %Y",tm);
  sprintf(__seconds,"%d",caltime);
  
  history_date = string("history") + "\t" +
    string("date") + "\t" +
    string (__date);
  
  history_utc_seconds = string("history") + "\t" +
    string("UTC_seconds") + "\t" + 
    string (__seconds);

}

void MultiChannelIcsInfo::addProperty(const vector<string> &a)
{
  if (a.empty()) return;
  string s = a[0];
  for(int i=1; i<a.size(); i++)
    s += "\t" + a[i];
  other_properties.push_back(s);
}


void MultiChannelIcsInfo::annotate(const string &_annotation)
{
  annotation = string("history\tAnnotation\t") + _annotation;
}

void MultiChannelIcsInfo::setScope(const string &_scope)
{
  scope = string("history\tScope\t") + _scope;
}

void MultiChannelIcsInfo::setSample(const string &dish_type, const string &cell_type,
	       const string &media_name, float media_vol,
	       float media_con, const string &oil_type)
{
  sample = string("Sample") + "\t" + dish_type + "\t" + cell_type + "\t" + media_name;
  char buf[256];
  sprintf(buf,"%f\0", media_vol);
  sample += string("\t") + string(buf);
  sprintf(buf,"%f\0", media_con);
  sample += string("\t") + string(buf) + "\t" + oil_type;
}


const vector<string> &MultiChannelIcsInfo::channelNames() const
{
  return channel;
}


ostream& image_io::operator<< (ostream &stream, const MultiChannelIcsInfo & info)
{
  // Write parameters
  const int nch = min( info.param_ex_filter.size(), info.param_exp_time.size() );
  if (nch==0) {
    stream << "parameter\tex_filter\tUNDEFINED" << endl;
    stream << "parameter\texp_time\tUNDEFINED" << endl;
  } else 
    for(int i=0; i<nch; i++) {
      stream << info.param_ex_filter[i] << endl;
      stream << info.param_exp_time[i] << endl;
    }
	       
  // Write history
  // Date
  if (!info.history_date.empty()) 
    stream << info.history_date << endl;
  else  
    stream << "history\tdate\tUNDEFINED" << endl;
  
  // UTC_seconds
  if (!info.history_utc_seconds.empty()) 
    stream << info.history_utc_seconds << endl;
  else
    stream << "history\tUTC_seconds\tUNDEFINED" << endl;
  
  // Scope
  if (!info.scope.empty()) 
    stream  << info.scope << endl;
  else 
    stream << "history\tScope\tUNDEFINED" << endl;

  // Sample
  if (!info.sample.empty()) 
    stream  << info.sample << endl;

  // Other properties
  for(int i=0; i<info.other_properties.size(); i++) 
    if (!info.other_properties[i].empty())
      stream << info.other_properties[i] << endl;

  // Annotation
  if (!info.annotation.empty()) 
    stream << info.annotation;
  else 
    stream << "history\tAnnotation\tNONE";

  return stream;
}




void image_io::writeICS(const MultiChannelIcsInfo &info, const Image &I)
{
  if (info.datasetName().empty()) return;

  using namespace fnp;
  string filename = clean(info.datasetName());

  string icsfile, idsfile;
  if ( (extension(filename)).compare("ics") ) {
    icsfile = string(filename)+".ics";
    idsfile = string(filename)+".ids";
  } else {    
    icsfile = string(filename);
    idsfile = root(filename)+".ids";
  }
  
  
  ofstream  outics(icsfile.c_str());  
  if(!outics) {
    cerr << "In image_io::writeICS" << endl;
    cerr << "   Can not open file" << icsfile << "\n";
    cerr << "   Ignored." << endl;
    return;
  }
  
  outics << "\t" << endl;
  outics << "ics_version\t1.0" << endl;
  outics << "filename\t" << tail(root(icsfile)) << endl;

  if (I.nz()>1) {
    outics << "layout\tparameters\t4" << endl;
    outics << "layout\torder\tbits\tx\ty\tz"  << endl;
    outics << "layout\tsizes\t"<< 8*I.pixelSize() << "\t" <<
      I.ny() << "\t" << I.nx() << "\t" << I.nz() << endl;
  } else {
    outics << "layout\tparameters\t3" << endl;
    outics << "layout\torder\tbits\tx\ty"  << endl;
    outics << "layout\tsizes\t"<< 8*I.pixelSize() << "\t" <<
      I.ny() << "\t" << I.nx() << endl;
  }

  outics << "layout\tcoordinates\tvideo" << endl;
  outics << "layout\tsignificant_bits\t" << 8*I.pixelSize() << endl;
  outics << "representation\tformat\t";
  if (I.format()!=FLOAT) outics << "integer" << endl;
  else outics<< "float" << endl;
  outics << "representation\tsign\tsigned"<< endl;
  outics << "representation\tcompression\tuncompressed" << endl;
  outics << "representation\tbyte_order\t2\t1"<< endl;
  if (I.nz()>1)outics << "representation\tSCIL_TYPE\tg3d" << endl;
  else  outics << "representation\tSCIL_TYPE\tg2d" << endl;

  outics << info << endl;

  outics.close();  
  
  ofstream    outimage(idsfile.c_str());
  if(!outimage) {
    cerr << "In image_io::writeICS" << endl;
    cout<< "   Can not open file" << idsfile << endl;
    cerr << "  In image_io::writeICS" << endl;
    return;
  }

  adjust_byte_order_for_write(I);  
  if (!outimage.write((char *) I.data(), I.npix()*I.pixelSize())) {
    cerr << "In image_io::writeICS" << endl;
    cerr << "   Error in writing " << idsfile<< endl;
  }
  adjust_byte_order_for_write(I);
  outimage.close();

}



void image_io::writePPM(const string &filename, const ByteImage &img)
{
  const int nx = img.nx();
  const int ny = img.ny();
  const int nz = img.nz();

  if (nz != 1 && nz != 3) {
    cerr << "image_io::writePPM -> img must be a grayscale or color image\n";
    return;
  }


  ofstream outimage(filename.c_str());  
  if(!outimage) {
    cerr << "Can not open file: " << filename << endl;
    return;
  }
  
  // Write header
  outimage << "P6" << endl;
  outimage << nx << " " << ny << endl;
  outimage << 255 << endl;

  const int nxy = nx*ny;
  const int nxy2 = nxy*2;
  const int len = nxy*3;
  byte *buf = new byte[len];

  byte *p = buf;
  byte *q = img.ptr();
  
  if (nz == 1)
    for(int i = 0; i < nxy; i++) {
      *(p++) = *q;
      *(p++) = *q;
      *(p++) = *(q++);
    } 
  else     
    for(int i = 0; i < nxy; i++) {
      *(p++) = *q;
      *(p++) = *(q+nxy);
      *(p++) = *(q+nxy2);
      q++;
    }
  
  // Write data
  outimage.write((char *) buf, len);
  outimage.close();
  
  delete []buf;
}

union uLongChar
{
 long num;
 char data[4];
};

union uShortChar
{
         unsigned short num;

        char data[2];

};
void extractShortDataFromBuffer(char buffer[],int lsb, int start, unsigned short &result)
{
  uShortChar val;

  if(lsb ==1)
  {
        val.data[0] = buffer[start+0];
        val.data[1] = buffer[start+1];
  }
        else
  {
        val.data[0] = buffer[start+3];
        val.data[1] = buffer[start+2];
  }
        result = val.num;
}

void extractLongDataFromBuffer(char buffer[],int lsb,int start, long & result)
{
  char buf[4];
  uLongChar val;
  if(lsb ==1)
 {
        val.data[0] = buffer[start+0];
        val.data[1] = buffer[start+1];
        val.data[2] = buffer[start+2];
        val.data[3] = buffer[start+3];
 }
 else
 {
        val.data[0] = buffer[start+3];
        val.data[1] = buffer[start+2];
        val.data[2] = buffer[start+1];
        val.data[3] = buffer[start+0];
 }
        result = val.num;
}



short returnStackCount(const string & filename){

 char buf[12];
 int bytesRead;
 long offsetIFD;
 int lsb;
 unsigned short stackCount=-1;
 FILE * imagePtr = fopen(filename.c_str(),"r");
 bool flag =1;
 int pos;
unsigned short numEntries;
unsigned short tiffTag;

 fread(buf,1,8,imagePtr);
 lsb =-1;
 if (buf[0]==0x49)
        lsb =1;
 else
        lsb=0;
 cout << "lsb"<<lsb<<endl;
 extractLongDataFromBuffer(buf,lsb,4,offsetIFD);
 
 int counter = 0;
 while(flag)
   {
     pos = fseek(imagePtr,offsetIFD,SEEK_SET);
     fread(buf,1,2,imagePtr);
     numEntries =0;
     extractShortDataFromBuffer(buf,lsb,0,numEntries);
     for(int i=0;i<numEntries;i++)
       {
	 fread(buf,1,12,imagePtr);
	 extractShortDataFromBuffer(buf,lsb,0,tiffTag);
	 if(tiffTag == TIFFTAG_METAMORPH)
	   {
	     extractShortDataFromBuffer(buf,lsb,4,stackCount);
	   }
       }
     fread(buf, 1, 4, imagePtr);
     extractLongDataFromBuffer(buf, lsb, 0,
			       offsetIFD);
     if(offsetIFD== 0) flag = 0;
     cout << counter++<<endl;
   }
 
 return stackCount;
 
}

